#include <popt.h>
#ifdef TRISH2
#  include <trish_dict.h>
#else
#  include <dictionary.h>
#endif
#include <string.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>

#include "cl_config.h"
#include "message.h"
#include "explore.h"
#include "directions.h"
#include "defaults.h"
#include "wchar_utils.h"

#include "pedantic_macros.h"
#include "mouchard.h"

int verbosity = VERBOSITY_NORMAL;
int severity = SEVERITY_ERROR;
static int np_len = 3;
static char *dict_n, *dict_fn;
static char *log_fn = NULL;
static int alvis = 0;
static int reannot = 0;

struct poptOption cl_options[] = {
  POPT_AUTOHELP
  { NULL , '\0' , POPT_ARG_INCLUDE_TABLE     , verb_opts , 0 , "Verbosity options:" , NULL },
  { CLO_LONG_CTX       , CLO_SHORT_CTX       , POPT_ARG_INT    , NULL      , CLO_SWITCH_CTX  , CLO_HELP_CTX       , "INT" },
  { CLO_LONG_DICT      , CLO_SHORT_DICT      , POPT_ARG_STRING , &dict_n   , CLO_SWITCH_DICT , CLO_HELP_DICT      , "NAME:PATH" },
  { CLO_LONG_NP_LENGTH , CLO_SHORT_NP_LENGTH , POPT_ARG_INT    , &np_len   , 0               , CLO_HELP_NP_LENGTH , "INT" },
  { CLO_LONG_LOG       , CLO_SHORT_LOG       , POPT_ARG_STRING , &log_fn   , 0               , CLO_HELP_LOG       , "PATH" },
  { CLO_LONG_ALVIS     , CLO_SHORT_ALVIS     , POPT_ARG_NONE   , &alvis    , 0               , CLO_HELP_ALVIS     , NULL },
  { CLO_LONG_REANNOT   , CLO_SHORT_REANNOT   , POPT_ARG_NONE   , &reannot  , 0               , CLO_HELP_REANNOT   , NULL },
  POPT_TABLEEND
};

int main(int argc, const char **argv) {
  int i;
  const char *in_fn;
  const char *class_fn;
  explore_t *ex = explore_new();
  FILE *f, *log_f = NULL;
  xmlDoc *class_doc, *in_doc;
  xmlXPathContext *xp_ctx;
  xmlXPathObject *xp_obj;
  int opth;
  poptContext context = poptGetContext(NULL, argc, argv, cl_options, 0);

  poptSetOtherOptionHelp(context, "CLASS FILE");

  while ((opth = poptGetNextOpt(context)) > 0) {
    switch (opth) {
    case CLO_SWITCH_QUIET:
      verbosity = VERBOSITY_QUIET;
      break;
    case CLO_SWITCH_VERBOSE:
      verbosity = VERBOSITY_VERBOSE;
      break;
    case CLO_SWITCH_LOCACE:
      verbosity = VERBOSITY_LOCACE;
      break;
    case CLO_SWITCH_STRICT:
      severity = SEVERITY_WARNING;
      break;
    case CLO_SWITCH_FORCE:
      severity = SEVERITY_FATAL;
      break;
    case CLO_SWITCH_DICT:
      dict_fn = strchr(dict_n, ':');
      if (dict_fn == NULL) {
	message(VERBOSITY_QUIET | SEVERITY_FATAL, "Dictionary argument should be in the form NAME:PATH");
      }
      *dict_fn = '\0';
      dict_fn++;
      if ((f = fopen(dict_fn, "r"))) {
#ifdef TRISH2
	tdict_t **tdp = bunch_claim(ex->ex_dict_list->data, 1);
	wchar_t *w = char2wcs(dict_n);
	*tdp = tdict_load(f);
	fclose(f);
	pmap_add(ex->ex_dict_list, w, tdp, 0);
	free(w);
#else
	dictionary_t *d = dictionary_load(f);
	fclose(f);
	association_insert(ex->ex_dict_list, dict_n, &d, 1);
#endif
      }
      else {
	message(VERBOSITY_QUIET | SEVERITY_ERROR, "Could not load dictionary '%s'", dict_fn);
      }
      break;
    case CLO_SWITCH_CTX:
      ex->ex_ctx = atoi(poptGetOptArg(context));
      break;
    }
  }
  if (opth != -1) {
    message(VERBOSITY_QUIET | SEVERITY_FATAL, "%s %s", poptStrerror(opth), poptBadOption(context, 0));
  }

  if (log_fn != NULL) {
    if (!(log_f = fopen(log_fn, "a"))) {
	message(VERBOSITY_QUIET | SEVERITY_ERROR, "Could not open log file '%s'", log_fn);
    }
  }

  if (poptPeekArg(context)) {
    class_fn =  poptGetArg(context);
    class_doc = xmlReadFile(class_fn, NULL, 0);
    if (class_doc == NULL) {
      message(VERBOSITY_QUIET | SEVERITY_FATAL, "Could not open classifier file %s", class_fn);
    }
  }
  else {
    message(VERBOSITY_QUIET | SEVERITY_FATAL, "You must provide a classifier file name");
  }
  if (poptPeekArg(context)) {
    if (alvis) {
      const char *alvis_fn = poptGetArg(context);
      xmlDoc *alvis_doc = xmlReadFile(alvis_fn, NULL, 0);
      xsltStylesheet *ss = NULL;
      int old_sub = xmlSubstituteEntitiesDefault(1);
      int old_led = xmlLoadExtDtdDefaultValue;

      xmlLoadExtDtdDefaultValue = 1;
      ss = xsltParseStylesheetFile(ALVIS_TO_FULL_XSLT);
      in_doc = xsltApplyStylesheet(ss, alvis_doc, NULL);

      xsltFreeStylesheet(ss);
      xmlFreeDoc(alvis_doc);
      xsltCleanupGlobals();
      xmlSubstituteEntitiesDefault(old_sub);
      xmlLoadExtDtdDefaultValue = old_led;
    }
    else {
      in_fn =  poptGetArg(context);
      in_doc = xmlReadFile(in_fn, NULL, 0);
      if (in_doc == NULL) {
	message(VERBOSITY_QUIET | SEVERITY_FATAL, "Could not open input file %s", in_fn);
      }
    }
  }
  else {
    message(VERBOSITY_QUIET | SEVERITY_FATAL, "You must provide an input file name");
  }
  poptFreeContext(context);

  explore_attributes(ex, class_doc);
  explore_dtree(ex, class_doc);

  xp_ctx = xmlXPathNewContext(in_doc);
  xp_obj = xmlXPathEvalExpression("//*[name() = 'ne' or (name() = 'word' and @category = 'N')]", xp_ctx);
  for (i = 0; i < xp_obj->nodesetval->nodeNr; i++) {
    noun_phrase_t *np = np_new(xp_obj->nodesetval->nodeTab[i], ex->ex_max_following_range, ex->ex_max_preceding_range);
    if (np) {
      int originally_ne = strcmp(xp_obj->nodesetval->nodeTab[i]->name, "word");
      do {
	const char *l = explore_get_label(ex, np, log_f);
	if (reannot) {
	  if (originally_ne) {
	    if ((l != NULL) && strcmp(l, "ne")) {
	      printf("-\t%s\n", xmlHasProp(xp_obj->nodesetval->nodeTab[i], "alvis-id")->children->content);
	    }
	    else {
	      printf("*\t%s\t%s\n", xmlHasProp(xp_obj->nodesetval->nodeTab[i], "alvis-id")->children->content, np->np_canonical ? np->np_canonical : "");
	    }
	    originally_ne = 0;
	  }
	  else {
	    if ((l == NULL) || strcmp(l, "not_ne")) {
	      int i;
	      printf("+\t%s\t%s", np->np_form, np->np_canonical ? np->np_canonical : "");
	      for (i = np->np_length; i; i--) {
		printf("\t%s", xmlHasProp(np->np_nodes[i - 1], "alvis-id")->children->content);
	      }
	      printf("\n");
	    }
	  }
	}
	else {
	  printf("%s\t%s\n", np->np_form, l);
	}
      } while (np_extend(np, np_len));
      np_free(np);
    }
  }
  xmlXPathFreeObject(xp_obj);
  xmlXPathFreeContext(xp_ctx);
  xmlFreeDoc(in_doc);
  xmlFreeDoc(class_doc);
  xmlCleanupParser();
  explore_free(ex);

  return 0;
}
