#include "explore.h"
#include "directions.h"
#include "eval_attribute.h"

#include <libxml/xpath.h>
#include <string.h>

#ifdef TRISH2
#  include <trish_dict.h>
#else
#  include <dictionary.h>
#endif

#include "pedantic_macros.h"
#include "mouchard.h"

explore_t *explore_new() {
  explore_t *ex = malloc(sizeof(explore_t));

  ex->ex_ctx = 0;
#ifdef TRISH2
  ex->ex_dict_list = patricia_new(DEFAULT_DICT_LIST_NB, DEFAULT_DICT_LIST_SB, bunch_new(sizeof(tdict_t *), 2));
#else
  ex->ex_dict_list = association_new(sizeof(dictionary_t *), 2);
#endif
  ex->ex_attr_list = NULL;
  ex->ex_max_following_range = 0;
  ex->ex_max_preceding_range = 0;

  return ex;
}

void explore_free(explore_t *ex) {
#ifdef TRISH2
  unsigned int i, j;
  for (i = 0; i < ex->ex_dict_list->data->block_number; i++)
    for (j = 0; j < ex->ex_dict_list->data->block_size - ex->ex_dict_list->data->blocks[i]->unclaimed; j++) {
      tdict_t **tdp = ((tdict_t **)ex->ex_dict_list->data->blocks[i]->data) + j;
      tdict_free(*tdp);
    }
  patricia_free(ex->ex_dict_list);
#else
  bunch_element_ref_t i;
  for (i = 1; i < ex->ex_dict_list->values->next_ref; i++) {
    dictionary_free(*BUNCH_ELT(dictionary_t *, ex->ex_dict_list->values, i));
  }
  association_free(ex->ex_dict_list);
#endif
  alist_free(ex->ex_attr_list);
  dtree_free(ex->ex_dt);
  free(ex);
}

void explore_attributes(explore_t *ex, xmlDoc *doc) {
  int i;
  xmlXPathContext *ctx = xmlXPathNewContext(doc);
  xmlXPathObject *obj = xmlXPathEvalExpression("/RenBio-rules/attribute-list/attribute/*", ctx);

  if (obj == NULL)
    return;

  ex->ex_attr_list = alist_new(obj->nodesetval->nodeNr);
  for (i = 0; i < obj->nodesetval->nodeNr; i++) {
    xmlNode *n;
    xmlAttr *id;

    n = obj->nodesetval->nodeTab[i];
    id = xmlHasProp(n->parent, "id");

    if (!strcmp(n->name, "re")) {
      alist_add_re(ex->ex_attr_list, id->children->content, n->children->content);
    }
    else if (!strcmp(n->name, "dict")) {
      xmlAttr *opa = xmlHasProp(n, "op");
      int op = ATTR_DICT_OP_EQUALS;
      if (!strcmp(opa->children->content,"equals")) {
	op = ATTR_DICT_OP_EQUALS;
      }
      else if (!strcmp(opa->children->content,"contains")) {
	op = ATTR_DICT_OP_CONTAINS;
      }
      alist_add_dict(ex->ex_attr_list, id->children->content, n->children->content, op);
    }
    else if (!strcmp(n->name, "length")) {
      alist_add_length(ex->ex_attr_list, id->children->content);
    }
    else if (!strcmp(n->name, "signal")) {
      xmlAttr *ra = xmlHasProp(n, "range");
      xmlAttr *da = xmlHasProp(n, "direction");
      int dir = DIRECTION_FOLLOWING;
      int range = atoi(ra->children->content);
      if (!strcmp(da->children->content, "following")) {
	dir = DIRECTION_FOLLOWING;
	if (range > ex->ex_max_following_range)
	  ex->ex_max_following_range = range;	
      }
      else if (!strcmp(da->children->content, "preceding")) {
	dir = DIRECTION_PRECEDING;
	if (range > ex->ex_max_preceding_range)
	  ex->ex_max_preceding_range = range;
      }
      alist_add_signal(ex->ex_attr_list, id->children->content, n->children->content, range, dir);
    }
    else if (!strcmp(n->name, "pos")) {
      xmlAttr *ra = xmlHasProp(n, "range");
      xmlAttr *da = xmlHasProp(n, "direction");
      int dir = DIRECTION_FOLLOWING;
      int range = atoi(ra->children->content);
      if (!strcmp(da->children->content, "following")) {
	dir = DIRECTION_FOLLOWING;
	if (range > ex->ex_max_following_range)
	  ex->ex_max_following_range = range;
      }
      else if (!strcmp(da->children->content, "preceding")) {
	dir = DIRECTION_PRECEDING;
	if (range > ex->ex_max_preceding_range)
	  ex->ex_max_preceding_range = range;
      }
      alist_add_pos(ex->ex_attr_list, id->children->content, range, dir);
    }
    else if (!strcmp(n->name, "context")) {
      alist_add_context(ex->ex_attr_list, id->children->content);
    }
    else if (!strcmp(n->name, "label")) {
      alist_add_label(ex->ex_attr_list, id->children->content);
    }
  }

  xmlXPathFreeObject(obj);
  xmlXPathFreeContext(ctx);
}

void explore_dtree_rec(explore_t *ex, dtree_t *dt, xmlNode *n) {
  xmlAttr *a = xmlHasProp(n, "attribute");

  dtree_attribute(dt, ex->ex_attr_list, a->children->content);
  for (n = n->children; n; n = n->next) {
    if ((n->type == XML_ELEMENT_NODE) && (!strcmp(n->name, "branch"))) {
      int op = 0;
      dtree_t *new;
      xmlNode *m;

      a = xmlHasProp(n, "op");
      if (!strcmp(a->children->content, "eq")) {
	op = DTREE_OP_EQ;
      }
      else if (!strcmp(a->children->content, "le")) {
	op = DTREE_OP_LE;
      }
      else if (!strcmp(a->children->content, "gt")) {
	op = DTREE_OP_GT;
      }
      a = xmlHasProp(n , "value");
      new = dtree_new(op, atoi(a->children->content), a->children->content);
      dtree_else(dt, new);

      for (m = n->children; m && strcmp(m->name, "node"); m = m->next);
      if (m) {
	explore_dtree_rec(ex, new, m);
      }
    }
  }
}

void explore_dtree(explore_t *ex, xmlDoc *doc) {
  xmlXPathContext *ctx = xmlXPathNewContext(doc);
  xmlXPathObject *obj = xmlXPathEvalExpression("/RenBio-rules/node", ctx);

  ex->ex_dt = dtree_new(0, 0, NULL);
  explore_dtree_rec(ex, ex->ex_dt, obj->nodesetval->nodeTab[0]);

  xmlXPathFreeObject(obj);
  xmlXPathFreeContext(ctx);
}

#ifdef TRISH2
const char *explore_get_label(explore_t *ex, noun_phrase_t *np, FILE *log) {
  dtree_t *dt = ex->ex_dt;

  if ((ex->ex_attr_list == NULL) || (ex->ex_dt == NULL)) {
    return NULL;
  }

  if ((np->np_following == NULL) || (np->np_preceding == NULL)) {
    return NULL;
  }

  if (log)
    fprintf(log, "Getting label for '%s'\n", np->np_form);

  while (dt->dtree_attribute->attr_type != ATTRIBUTE_TYPE_LABEL) {
    int cont = 1;
    eval_attribute_t *ea = eval_attribute(ex->ex_dict_list, ex->ex_ctx, dt->dtree_attribute, np);
    if (ea->ea_ival && ea->ea_sval)
      np->np_canonical = ea->ea_sval;
    if (log)
      fprintf(log, "%s ", dt->dtree_attribute->attr_name);
    dt = dt->dtree_test;
    while (cont) {
      if (dt == NULL)
	return NULL;
      switch (ea->ea_type) {
      case EVAL_ATTRIBUTE_TYPE_INT:
	switch (dt->dtree_op) {
	case DTREE_OP_EQ:
	  if (ea->ea_ival == dt->dtree_ival) {
	    cont = 0;
	    if (log)
	      fprintf(log, "(%d) = %d\n", ea->ea_ival, dt->dtree_ival);
	  }
	  break;
	case DTREE_OP_LE:
	  if (ea->ea_ival <= dt->dtree_ival) {
	    cont = 0;
	    if (log)
	      fprintf(log, "(%d) <= %d\n", ea->ea_ival, dt->dtree_ival);
	  }
	  break;
	case DTREE_OP_GT:
	  if (ea->ea_ival > dt->dtree_ival) {
	    cont = 0;
	    if (log)
	      fprintf(log, "(%d) > %d\n", ea->ea_ival, dt->dtree_ival);
	  }
	  break;
	default:
	  if (log)
	    fprintf(log, "(%d) ? %d\n", ea->ea_ival, dt->dtree_ival);
	}
	break;
      case EVAL_ATTRIBUTE_TYPE_STRING:
	switch (dt->dtree_op) {
	case DTREE_OP_EQ:
	  if (ea->ea_sval == NULL)
	    break;
	  if (!strcmp(ea->ea_sval, dt->dtree_sval)) {
	    cont = 0;
	    if (log)
	      fprintf(log, "('%s') = '%s'\n", ea->ea_sval, dt->dtree_sval);
	  }
	  break;
	case DTREE_OP_LE:
	  if (ea->ea_sval == NULL) {
	    cont = 0;
	    break;
	  }
	  if (strcmp(ea->ea_sval, dt->dtree_sval) <= 0) {
	    cont = 0;
	    if (log)
	      fprintf(log, "('%s') <= '%s'\n", ea->ea_sval, dt->dtree_sval);
	  }
	  break;
	case DTREE_OP_GT:
	  if (ea->ea_sval == NULL)
	    break;
	  if (strcmp(ea->ea_sval, dt->dtree_sval) > 0) {
	    cont = 0;
	    if (log)
	      fprintf(log, "('%s') > '%s'\n", ea->ea_sval, dt->dtree_sval);
	  }
	  break;
	default:
	  if (log)
	    fprintf(log, "('%s') ? '%s'\n", ea->ea_sval, dt->dtree_sval);
	}
	break;
      default:
	if (log) {
	  fprintf(log, "unknown answer\n");
	}
      }
      if (cont) {
	dt = dt->dtree_else;
      }
    }
    free(ea);
  }

  if (log)
    fprintf(log, "Conclusion: %s\n\n", dt->dtree_attribute->attr_name);
  return dt->dtree_attribute->attr_name;

#else

const char *explore_get_label(explore_t *ex, noun_phrase_t *np, FILE *log) {
  dtree_t *dt = ex->ex_dt;

  if ((ex->ex_attr_list == NULL) || (ex->ex_dt == NULL)) {
    return NULL;
  }

  if ((np->np_following == NULL) || (np->np_preceding == NULL)) {
    return NULL;
  }

  if (log)
    fprintf(log, "Getting label for '%s'\n", np->np_form);

  while (BUNCH_ELT(attribute_t, ex->ex_attr_list->values, dt->dtree_attribute)->attr_type != ATTRIBUTE_TYPE_LABEL) {
    int cont = 1;
    eval_attribute_t *ea = eval_attribute_ref(ex->ex_dict_list, ex->ex_ctx, ex->ex_attr_list, dt->dtree_attribute, np);
    if (ea->ea_ival && ea->ea_sval)
      np->np_canonical = ea->ea_sval;
    if (log)
      fprintf(log, "%s ", BUNCH_ELT(attribute_t, ex->ex_attr_list->values, dt->dtree_attribute)->attr_name);
    dt = dt->dtree_test;
    while (cont) {
      switch (ea->ea_type) {
      case EVAL_ATTRIBUTE_TYPE_INT:
	switch (dt->dtree_op) {
	case DTREE_OP_EQ:
	  if (ea->ea_ival == dt->dtree_ival) {
	    cont = 0;
	    if (log)
	      fprintf(log, "(%d) = %d\n", ea->ea_ival, dt->dtree_ival);
	  }
	  break;
	case DTREE_OP_LE:
	  if (ea->ea_ival <= dt->dtree_ival) {
	    cont = 0;
	    if (log)
	      fprintf(log, "(%d) <= %d\n", ea->ea_ival, dt->dtree_ival);
	  }
	  break;
	case DTREE_OP_GT:
	  if (ea->ea_ival > dt->dtree_ival) {
	    cont = 0;
	    if (log)
	      fprintf(log, "(%d) > %d\n", ea->ea_ival, dt->dtree_ival);
	  }
	  break;
	default:
	  if (log)
	    fprintf(log, "(%d) ? %d\n", ea->ea_ival, dt->dtree_ival);
	}
	break;
      case EVAL_ATTRIBUTE_TYPE_STRING:
	switch (dt->dtree_op) {
	case DTREE_OP_EQ:
	  if (!strcmp(ea->ea_sval, dt->dtree_sval)) {
	    cont = 0;
	    if (log)
	      fprintf(log, "('%s') = '%s'\n", ea->ea_sval, dt->dtree_sval);
	  }
	  break;
	case DTREE_OP_LE:
	  if (strcmp(ea->ea_sval, dt->dtree_sval) <= 0) {
	    cont = 0;
	    if (log)
	      fprintf(log, "('%s') <= '%s'\n", ea->ea_sval, dt->dtree_sval);
	  }
	  break;
	case DTREE_OP_GT:
	  if (strcmp(ea->ea_sval, dt->dtree_sval) > 0) {
	    cont = 0;
	    if (log)
	      fprintf(log, "('%s') > '%s'\n", ea->ea_sval, dt->dtree_sval);
	  }
	  break;
	default:
	  if (log)
	    fprintf(log, "('%s') ? '%s'\n", ea->ea_sval, dt->dtree_sval);
	}
	break;
      default:
	if (log) {
	  fprintf(log, "unknown answer\n");
	}
      }
      if (cont) {
	dt = dt->dtree_else;
      }
    }
    free(ea);
  }

  if (log)
    fprintf(log, "Conclusion: %s\n\n", BUNCH_ELT(attribute_t, ex->ex_attr_list->values, dt->dtree_attribute)->attr_name);
  return BUNCH_ELT(attribute_t, ex->ex_attr_list->values, dt->dtree_attribute)->attr_name;
#endif
}
