#include <popt.h>
#ifdef TRISH2
#  include <trish_dict.h>
#else
#  include <dictionary.h>
#endif
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <string.h>

#include "explore.h"
#include "eval_attribute.h"
#include "cl_config.h"
#include "message.h"
#include "wchar_utils.h"

#include "pedantic_macros.h"
#include "mouchard.h"

#define XPATH_QUERY_ALL "//*[(name() = 'ne' and @type = '%s') or (name() = 'word' and @category = 'N')]"

int verbosity = VERBOSITY_NORMAL;
int severity = SEVERITY_ERROR;
static int np_len = 3;
static char *dict_n, *dict_fn;
static char *type;

struct poptOption cl_options[] = {
  POPT_AUTOHELP
  { NULL , '\0' , POPT_ARG_INCLUDE_TABLE     , verb_opts , 0 , "Verbosity options:" , NULL },
  { CLO_LONG_CTX       , CLO_SHORT_CTX       , POPT_ARG_INT    , NULL      , CLO_SWITCH_CTX  , CLO_HELP_CTX       , "INT" },
  { CLO_LONG_DICT      , CLO_SHORT_DICT      , POPT_ARG_STRING , &dict_n   , CLO_SWITCH_DICT , CLO_HELP_DICT      , "NAME:PATH" },
  { CLO_LONG_NP_LENGTH , CLO_SHORT_NP_LENGTH , POPT_ARG_INT    , &np_len   , 0               , CLO_HELP_NP_LENGTH , "INT" },
  { CLO_LONG_TYPE      , CLO_SHORT_TYPE      , POPT_ARG_STRING , &type     , 0               , CLO_HELP_TYPE      , "TYPE" },
  POPT_TABLEEND
};

int main(int argc, const char **argv) {
  const char *in_fn;
  const char *class_fn;
  xmlDoc *class_doc;
  int opth;
  unsigned int i, j, k;
  FILE *f;
  explore_t *ex = explore_new();
  poptContext context = poptGetContext(NULL, argc, argv, cl_options, 0);

  poptSetOtherOptionHelp(context, "CLASS FILE [FILE...]");

  while ((opth = poptGetNextOpt(context)) > 0) {
    switch (opth) {
    case CLO_SWITCH_QUIET:
      verbosity = VERBOSITY_QUIET;
      break;
    case CLO_SWITCH_VERBOSE:
      verbosity = VERBOSITY_VERBOSE;
      break;
    case CLO_SWITCH_LOCACE:
      verbosity = VERBOSITY_LOCACE;
      break;
    case CLO_SWITCH_STRICT:
      severity = SEVERITY_WARNING;
      break;
    case CLO_SWITCH_FORCE:
      severity = SEVERITY_FATAL;
      break;
    case CLO_SWITCH_DICT:
      dict_fn = strchr(dict_n, ':');
      if (dict_fn == NULL) {
	message(VERBOSITY_QUIET | SEVERITY_FATAL, "Dictionary argument should be in the form NAME:PATH");
      }
      *dict_fn = '\0';
      dict_fn++;
      if ((f = fopen(dict_fn, "r"))) {
#ifdef TRISH2
	tdict_t **tdp = bunch_claim(ex->ex_dict_list->data, 1);
	wchar_t *w = char2wcs(dict_n);
	*tdp = tdict_load(f);
	fclose(f);
	pmap_add(ex->ex_dict_list, w, tdp, 0);
	free(w);
#else
	dictionary_t *d = dictionary_load(f);
	fclose(f);
	association_insert(ex->ex_dict_list, dict_n, &d, 1);
#endif
      }
      else {
	message(VERBOSITY_QUIET | SEVERITY_ERROR, "Could not load dictionary '%s'", dict_fn);
      }
      break;
    case CLO_SWITCH_CTX:
      ex->ex_ctx = atoi(poptGetOptArg(context));
      break;
    }
  }
  if (opth != -1) {
    message(VERBOSITY_QUIET | SEVERITY_FATAL, "%s %s", poptStrerror(opth), poptBadOption(context, 0));
  }

  if (poptPeekArg(context)) {
    class_fn =  poptGetArg(context);
    class_doc = xmlReadFile(class_fn, NULL, 0);
    if (class_doc == NULL) {
      message(VERBOSITY_QUIET | SEVERITY_FATAL, "Could not open classifier file %s", class_fn);
    }
  }
  else {
    message(VERBOSITY_QUIET | SEVERITY_FATAL, "You must provide a classifier file name");
  }

  printf("@RELATION ne_or_not\n\n");

  explore_attributes(ex, class_doc);

#ifdef TRISH2
  for (i = 0; i < ex->ex_attr_list->data->block_number; i++)
    for (j = 0; j < ex->ex_attr_list->data->block_size - ex->ex_attr_list->data->blocks[i]->unclaimed; j++) {
      attribute_t *attr = ((attribute_t *)ex->ex_attr_list->data->blocks[i]->data) + j;
      switch (attr->attr_type) {
      case ATTRIBUTE_TYPE_LENGTH:
	printf("@ATTRIBUTE %s NUMERIC\n", attr->attr_name);
	break;
      case ATTRIBUTE_TYPE_POS:
	printf("@ATTRIBUTE %s {J,N,PP,RB,V,O}\n", attr->attr_name);
	break;
      case ATTRIBUTE_TYPE_LABEL:
	break;
      default:
	printf("@ATTRIBUTE %s {0,1}\n", attr->attr_name);
	break;
      }
    }
#else
  for (i = 1; i < ex->ex_attr_list->values->next_ref; i++) {
    switch (BUNCH_ELT(attribute_t, ex->ex_attr_list->values, i)->attr_type) {
    case ATTRIBUTE_TYPE_LENGTH:
      printf("@ATTRIBUTE %s NUMERIC\n", BUNCH_ELT(attribute_t, ex->ex_attr_list->values, i)->attr_name);
      break;
    case ATTRIBUTE_TYPE_POS:
      printf("@ATTRIBUTE %s {J,N,PP,RB,V,O}\n", BUNCH_ELT(attribute_t, ex->ex_attr_list->values, i)->attr_name);
      break;
    case ATTRIBUTE_TYPE_LABEL:
      break;
    default:
      printf("@ATTRIBUTE %s {0,1}\n", BUNCH_ELT(attribute_t, ex->ex_attr_list->values, i)->attr_name);
      break;
    }
  }
#endif
  printf("@ATTRIBUTE class {ne,not_ne}\n\n@DATA\n");

  if (poptPeekArg(context)) {
    while ((in_fn =  poptGetArg(context))) {
      xmlDoc *in_doc = xmlReadFile(in_fn, NULL, 0);
      if (in_doc == NULL) {
	message(VERBOSITY_QUIET | SEVERITY_ERROR, "Could not open input file %s", in_fn);
      }
      else {
	xmlXPathContext *xp_ctx = xmlXPathNewContext(in_doc);
	xmlXPathObject *xp_obj = xmlXPathEvalExpression("//*[(name() = 'ne') or (name() = 'word' and @category = 'N')]", xp_ctx);

	printf("%% %s\n", in_fn);
	for (k = 0; k < xp_obj->nodesetval->nodeNr; k++) {
	  xmlAttr *a = xmlHasProp(xp_obj->nodesetval->nodeTab[k], "type");
	  noun_phrase_t *np = np_new(xp_obj->nodesetval->nodeTab[k], ex->ex_max_following_range, ex->ex_max_preceding_range);
	  if (np) {
	    do {
	      printf("%% %s\n", np->np_form);
#ifdef TRISH2
	      for (i = 0; i < ex->ex_attr_list->data->block_number; i++)
		for (j = 0; j < ex->ex_attr_list->data->block_size - ex->ex_attr_list->data->blocks[i]->unclaimed; j++) {
		  attribute_t *attr = ((attribute_t *)ex->ex_attr_list->data->blocks[i]->data) + j;
		  if (attr->attr_type != ATTRIBUTE_TYPE_LABEL) {
		    eval_attribute_t *ea = eval_attribute(ex->ex_dict_list, ex->ex_ctx, attr, np);
		    if (ea->ea_type == EVAL_ATTRIBUTE_TYPE_INT) {
		      printf("%i,", ea->ea_ival);
		    }
		    else {
		      printf("%s,", (ea->ea_sval == NULL) ? "?" : ea->ea_sval);
		    }
		    free(ea);
		  }
		}
#else
	      for (i = 1; i < ex->ex_attr_list->values->next_ref; i++) {
		eval_attribute_t *ea = eval_attribute_ref(ex->ex_dict_list, ex->ex_ctx, ex->ex_attr_list, i, np);
		if (ea->ea_type == EVAL_ATTRIBUTE_TYPE_INT) {
		  printf("%i,", ea->ea_ival);
		}
		else {
		  printf("%s,", ea->ea_sval);
		}
		free(ea);
	      }
#endif
	      printf("%s\n", (strcmp(xp_obj->nodesetval->nodeTab[k]->name, "ne") || a == NULL || strcmp(a->children->content, type)) ? "not_ne" : "ne");
	    } while (np_extend(np, np_len));
	    np_free(np);
	  }
	}
	xmlXPathFreeObject(xp_obj);
	xmlXPathFreeContext(xp_ctx);
	xmlFreeDoc(in_doc);
      }
    }
  }
  else {
    message(VERBOSITY_QUIET | SEVERITY_FATAL, "You must provide an input file name");
  }

  poptFreeContext(context);
  xmlFreeDoc(class_doc);
  xmlCleanupParser();
  explore_free(ex);

  return 0;
}
